package laboratoire4.pawns;

import laboratoire4.Action;
import laboratoire4.IPawn;
import laboratoire4.Player;
import laboratoire4.game.Game;
import laboratoire4.game.GameUtils;

import java.awt.*;
import java.util.List;
import java.util.Queue;
import java.util.*;

public class PawnUtils {
    public static int distanceFromGoal(IPawn pawn) {
        return Math.abs(pawn.getPlayer().getGoal() - pawn.getRow());
    }

    public static int distanceFromHome(IPawn pawn) {
        return Math.abs(pawn.getRow() - pawn.getPlayer().getHome());
    }

    public static boolean areSamePlayers(IPawn a, IPawn b) {
        return a != null && b != null && a.getPlayer() == b.getPlayer();
    }

    public static boolean hasEasyWinPath(IPawn[][] board, IPawn initialPawn) {
        Queue<Point> positionsToVisit = new ArrayDeque<>();
        positionsToVisit.add(new Point(initialPawn.getRow(), initialPawn.getCol()));

        while (!positionsToVisit.isEmpty()) {
            Point pos = positionsToVisit.poll();

            for (PawnMovement movement : PawnMovement.values()) {
                if (!initialPawn.isMoveValid(board, movement, pos.x, pos.y, false)) {
                    continue;
                }

                int nextRow = pos.x + initialPawn.getDirection();
                int nextCol = pos.y + movement.getMove();

                IPawn pawn = board[nextRow][nextCol];
                if (pawn != null && pawn.getPlayer() != initialPawn.getPlayer() && pawn.isMoveValid(board, movement.getOpposite())) {
                    //  On peut se faire capturer en allant là !
                    continue;
                }

                if (nextRow == initialPawn.getPlayer().getGoal()) {
                    return true;
                }

                positionsToVisit.add(new Point(nextRow, nextCol));
            }
        }

        return false;
    }

    public static boolean canBeCaptured(Game game, IPawn pawn) {
        for (PawnMovement movement : PawnMovement.values()) {
            int nextRow = pawn.getRow() + pawn.getDirection();
            int nextCol = pawn.getCol() + movement.getMove();

            if (nextRow < 0 || nextRow > 7 || nextCol < 0 || nextCol > 7) {
                continue;
            }

            IPawn nearPawn = game.getBoard()[nextRow][nextCol];
            if (nearPawn == null || PawnUtils.areSamePlayers(pawn, nearPawn)) {
                continue;
            }

            if (nearPawn.isMoveValid(game.getBoard(), movement.getOpposite())) {
                return true;
            }
        }

        return false;
    }

    public static boolean canBeCaptured(Game game, int row, int col, Player player) {
        for (PawnMovement movement : PawnMovement.values()) {
            int nextRow = row + player.getDirection();
            int nextCol = col + movement.getMove();

            if (nextRow < 0 || nextRow > 7 || nextCol < 0 || nextCol > 7) {
                continue;
            }

            IPawn nearPawn = game.getBoard()[nextRow][nextCol];
            if (nearPawn == null || player == nearPawn.getPlayer()) {
                continue;
            }

            if (nearPawn.isMoveValid(game.getBoard(), movement.getOpposite(), true)) {
                return true;
            }
        }

        return false;
    }

    public static boolean canCapture(Game game, IPawn max, IPawn min) {
        int rowDistance = Math.abs(min.getRow() - max.getRow());
        int colDistance = Math.abs(min.getCol() - max.getCol());

        if (colDistance > rowDistance) {
            // Le pion ne pourra jamais être atteint
            return false;
        }

        return PawnUtils.canCapture(game, max, min, max.getRow(), max.getCol(), 0, rowDistance);
    }

    public static boolean canCapture(Game game, IPawn max, IPawn min, int row, int col, int depth, int maxDepth) {
        if (min.getRow() == row && min.getCol() == col) {
            return true;
        }

        if (depth >= maxDepth) {
            return false;
        }

        for (PawnMovement movement : PawnMovement.values()) {
            if (max.isMoveValid(game.getBoard(), movement, row, col, false)) {
                int nextRow = row + max.getDirection();
                int nextCol = col + movement.getMove();

                if (canCapture(game, max, min, nextRow, nextCol, depth + 1, maxDepth)) {
                    return true;
                }
            }
        }

        return false;
    }

    public static <T extends IPawn> Collection<Action<T>> getActions(Game game, boolean max) {
        return getActions(game, max, true);
    }

    public static <T extends IPawn> Collection<Action<T>> getActions(Game game, boolean max, boolean excludeDefense) {
        List<Action<T>> actions = new ArrayList<>();
        PawnMovement[] movements = PawnMovement.values();

        //noinspection unchecked
        Collection<T> pawns = (Collection<T>) (max ?
                GameUtils.getMaxPawns(game.getBoard(), game.getPlayer()) :
                GameUtils.getMinPawns(game.getBoard(), game.getPlayer()));

        for (T pawn : pawns) {
            if (excludeDefense && pawn.getRow() == pawn.getPlayer().getHome()) {
                int col = pawn.getCol();

                // Si possible, on ne bouge pas ces pushers, comme ça on a une défense d'urgence totale
                if (col == 1 || col == 2 || col == 5 || col == 6) {
                    continue;
                }
            }

            for (PawnMovement movement : movements) {
                if (pawn.isMoveValid(game.getBoard(), movement)) {
                    actions.add(new Action<>(pawn, movement));
                }
            }
        }

        if (excludeDefense && actions.isEmpty()) {
            return getActions(game, max, false);
        }

        Collections.shuffle(actions);
        return actions;
    }
}
