package laboratoire4.strategies;

import laboratoire4.Action;
import laboratoire4.IPawn;
import laboratoire4.game.Game;
import laboratoire4.game.GameUtils;
import laboratoire4.pawns.PawnUtils;
import laboratoire4.pawns.SimulatedPawn;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

public class DefenseStrategy extends MiniMaxStrategy {
    private static final int DEFENSE_DISTANCE_FROM_HOME = 3;
    private static final int PAWN_DEFENSE_DISTANCE = 3;

    private final Map<IPawn, Integer> dangerousPawns = new HashMap<>();

    @Override
    public EvaluationResult getNextMove(Game game) {
        for (IPawn pawn : GameUtils.getMinPawns(game.getBoard(), game.getPlayer())) {
            int distance = PawnUtils.distanceFromGoal(pawn);

            if (distance <= DEFENSE_DISTANCE_FROM_HOME) {
                dangerousPawns.put(pawn, DEFENSE_DISTANCE_FROM_HOME - distance + 1);
            }
        }

        if (dangerousPawns.isEmpty()) {
            return null;
        }

        return super.getNextMove(game);
//        EvaluationResult result = super.getNextMove(game);
//        if (Math.abs(result.getRow() - game.getPlayer().getHome()) > DEFENSE_DISTANCE_FROM_HOME) {
//            return null;
//        }
//
//        return result;
    }

    @Override
    public int getWeight(Game game) {
        Collection<IPawn> minPawns = GameUtils.getMinPawns(game.getBoard(), game.getPlayer());
        Collection<IPawn> maxPawns = GameUtils.getMaxPawns(game.getBoard(), game.getPlayer());
        int maxWeight = 0;

        for (IPawn pawn : minPawns) {
            int weight = 0;

            int distanceFromGoal = PawnUtils.distanceFromGoal(pawn);
            if (distanceFromGoal <= 2) {
                if (PawnUtils.hasEasyWinPath(game.getBoard(), pawn)) {
                    if (maxPawns.stream().noneMatch(p -> PawnUtils.canCapture(game, p, pawn))) {
                        // Le pion ne peut pas être arrêté, on doit attaquer !
                        return 0;
                    } else {
                        return WEIGHT_MAX;
                    }
                } else {
                    weight = 5;
                }
            }

            if (distanceFromGoal > 2 && distanceFromGoal <= DEFENSE_DISTANCE_FROM_HOME) {
                weight = 2;
            }

            if (weight > maxWeight) {
                maxWeight = weight;
            }
        }

        return maxWeight;
    }

    @Override
    protected int evaluateSimulation() {
        int score = 0;

        Collection<SimulatedPawn> maxPawns = getMaxPawns();
        Collection<SimulatedPawn> minPawns = getMinPawns();

        for (SimulatedPawn pawn : minPawns) {
            if (pawn.getRow() == pawn.getPlayer().getGoal()) {
                return Integer.MIN_VALUE;
            }
        }

        for (IPawn dangerousPawn : dangerousPawns.keySet()) {
            Optional<SimulatedPawn> simulated = GameUtils.getMinPawnsAsStream(game.getBoard(), game.getPlayer())
                    .filter(p -> p.getOriginalRow() == dangerousPawn.getRow() && p.getOriginalCol() == dangerousPawn.getCol())
                    .findFirst();

            if (simulated.isEmpty()) {
                // Le pion a été capturé !
                score += Math.pow(dangerousPawns.get(dangerousPawn), 3);
            } else {
                // On est toujours en danger
                int distance = PawnUtils.distanceFromGoal(simulated.get());
                score -= Math.pow(DEFENSE_DISTANCE_FROM_HOME - distance + 1, 3);

                for (SimulatedPawn pawn : maxPawns) {
                    score += PAWN_DEFENSE_DISTANCE - distanceFromCapture(pawn, simulated.get()) * 5;
                }
            }
        }

        for (SimulatedPawn pawn : maxPawns) {
            if (PawnUtils.canBeCaptured(game, pawn)) {
                score -= 5;
            }

            // On préfère les groupes de 2 pushers, puisqu'ils ont une meilleure surface de défense
            if (!pawn.isPusher()) {
                continue;
            }

            int col = pawn.getCol();
            int row = pawn.getRow();

            if (col > 0 && PawnUtils.areSamePlayers(pawn, game.getBoard()[col - 1][row])) {
                score += 5;
            }
            if (col < 7 && PawnUtils.areSamePlayers(pawn, game.getBoard()[col + 1][row])) {
                score += 5;
            }
        }

        return score;
    }

    @Override
    protected Collection<Action<IPawn>> getActions(boolean max) {
        Collection<SimulatedPawn> pawns = max ? getMaxPawns() : getMinPawns();

        long defensivePawnCount = pawns.stream()
                .filter(IPawn::isPusher)
                .filter(p -> p.getRow() != p.getPlayer().getHome())
                .filter(p -> PawnUtils.distanceFromHome(p) <= DEFENSE_DISTANCE_FROM_HOME)
                .count();

        return PawnUtils.getActions(game, max, defensivePawnCount > 0);
    }

    private int distanceFromCapture(IPawn max, IPawn min) {
        int rowDistance = Math.abs(min.getRow() - max.getRow());
        int colDistance = Math.abs(min.getCol() - max.getCol());

        if (colDistance > rowDistance) {
            // Le pion ne pourra jamais être atteint
            return PAWN_DEFENSE_DISTANCE;
        }

        if (!PawnUtils.canCapture(game, max, min, max.getRow(), max.getCol(), 0, rowDistance)) {
            return PAWN_DEFENSE_DISTANCE;
        }

        return rowDistance;
    }


}
